
#include "E.h"

void
CommsSetup()
{
  char                s[1024];
  Atom                a;

  EDBUG(5, "CommsSetup");
  comms_win = XCreateSimpleWindow(disp, root.win, -100, -100, 5, 5, 0, 0, 0);
  Esnprintf(s, sizeof(s), "WINID %8x", (int)comms_win);
  a = XInternAtom(disp, "ENLIGHTENMENT_COMMS", False);
  XChangeProperty(disp, root.win, a, XA_STRING, 8, PropModeReplace, (unsigned char *)s,
		  strlen(s));
  a = XInternAtom(disp, "ENL_MSG", False);
  EDBUG_RETURN_;
}

void
CommsFindCommsWindow()
{
  unsigned char      *s;
  Atom                a, ar;
  unsigned long       num, after;
  int                 format;

  EDBUG(6, "CommsFindCommsWindow");
  a = XInternAtom(disp, "ENLIGHTENMENT_COMMS", True);
  if (a != None)
    {
      s = NULL;
      XGetWindowProperty(disp, root.win, a, 0, 14, False, AnyPropertyType, &ar,
			 &format, &num, &after, &s);
      if (s)
	sscanf((char *)s, "%*s %x", (unsigned int *)&comms_win);
      else
	(comms_win = 0);
      XFree(s);
    }
  EDBUG_RETURN_;
}

void
CommsSend(Client * c, char *s)
{
  char                ss[21];
  int                 i, j, k, len;
  XEvent              ev;
  Atom                a;

  EDBUG(5, "CommsSend");
  if ((!s) || (!c))
    EDBUG_RETURN_;
  len = strlen(s);
  a = XInternAtom(disp, "ENL_MSG", True);
  ev.xclient.type = ClientMessage;
  ev.xclient.serial = 0;
  ev.xclient.send_event = True;
  ev.xclient.window = c->win;
  ev.xclient.message_type = a;
  ev.xclient.format = 8;
  for (i = 0; i < len + 1; i += 12)
    {
      Esnprintf(ss, sizeof(ss), "%8x", (int)comms_win);
      for (j = 0; j < 12; j++)
	{
	  ss[8 + j] = s[i + j];
	  if (!s[i + j])
	    j = 12;
	}
      ss[20] = 0;
      for (k = 0; k < 20; k++)
	ev.xclient.data.b[k] = ss[k];
      XSendEvent(disp, c->win, False, 0, (XEvent *) & ev);
    }
  EDBUG_RETURN_;
}

char               *
CommsGet(Client ** c, XEvent * ev)
{
  char                s[13], s2[9], *msg, st[32];
  int                 i;
  Window              win;
  Client             *cl;

  EDBUG(5, "CommsGet");
  if ((!ev) || (!c))
    EDBUG_RETURN(NULL);
  if (ev->type != ClientMessage)
    EDBUG_RETURN(NULL);
  s[12] = 0;
  s2[8] = 0;
  msg = NULL;
  for (i = 0; i < 8; i++)
    s2[i] = ev->xclient.data.b[i];
  for (i = 0; i < 12; i++)
    s[i] = ev->xclient.data.b[i + 8];
  sscanf(s2, "%x", (int *)&win);
  cl = (Client *) FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_CLIENT);
  if (!cl)
    {
      cl = MakeClient(win);
      if (!cl)
	EDBUG_RETURN(NULL);
      Esnprintf(st, sizeof(st), "%8x", (int)win);
      cl->name = duplicate(st);
      AddItem((void *)cl, st, cl->win, LIST_TYPE_CLIENT);
      XSelectInput(disp, win, StructureNotifyMask | SubstructureNotifyMask);
    }
  if (cl->msg)
    {
      /* append text to end of msg */
      cl->msg = Erealloc(cl->msg, strlen(cl->msg) + strlen(s) + 1);
      if (!cl->msg)
	EDBUG_RETURN(NULL);
      strcat(cl->msg, s);
    }
  else
    {
      /* new msg */
      cl->msg = Emalloc(strlen(s) + 1);
      if (!cl->msg)
	EDBUG_RETURN(NULL);
      strcpy(cl->msg, s);
    }
  if (strlen(s) < 12)
    {
      msg = cl->msg;
      cl->msg = NULL;
      *c = cl;
    }
  EDBUG_RETURN(msg);
}

void
CommsBroadcast(char *s)
{
  char              **l;
  int                 num, i;
  Client             *c;

  EDBUG(5, "CommsBroadcast");
  l = ListItems(&num, LIST_TYPE_CLIENT);
  if (!s)
    EDBUG_RETURN_;
  for (i = 0; i < num; i++)
    {
      c = (Client *) FindItem(l[i], 0, LIST_FINDBY_NAME, LIST_TYPE_CLIENT);
      if (c)
	CommsSend(c, s);
    }
  freestrlist(l, num);
  EDBUG_RETURN_;
}

Client             *
MakeClient(Window win)
{
  Client             *c;

  EDBUG(6, "MakeClient");
  c = Emalloc(sizeof(Client));
  if (!c)
    EDBUG_RETURN(NULL);
  c->name = NULL;
  c->win = win;
  c->msg = NULL;
  c->clientname = NULL;
  c->version = NULL;
  c->author = NULL;
  c->email = NULL;
  c->web = NULL;
  c->address = NULL;
  c->info = NULL;
  c->pmap = 0;
  EDBUG_RETURN(c);
}

void
ListFreeClient(void *ptr)
{
  Client             *c;

  EDBUG(6, "ListFreeClient");
  c = (Client *) ptr;
  if (!c)
    EDBUG_RETURN_;
  if (c->name)
    Efree(c->name);
  if (c->msg)
    Efree(c->msg);
  if (c->clientname)
    Efree(c->clientname);
  if (c->version)
    Efree(c->version);
  if (c->author)
    Efree(c->author);
  if (c->email)
    Efree(c->email);
  if (c->web)
    Efree(c->web);
  if (c->address)
    Efree(c->address);
  if (c->info)
    Efree(c->info);
  Efree(c);
  EDBUG_RETURN_;
}

void
DeleteClient(Client * c)
{
  Client             *cc;

  EDBUG(6, "DeleteClient");
  cc = RemoveItem(NULL, c->win, LIST_FINDBY_ID, LIST_TYPE_CLIENT);
  ListFreeClient(cc);
  EDBUG_RETURN_;
}

void
HandleComms(XEvent * ev)
{
  Client             *c;
  char               *s, w[FILEPATH_LEN_MAX], w2[FILEPATH_LEN_MAX], *s1,
                     *s2;
  char                sunknown[] = "UNKNOWN";
  int                 unknown;

  EDBUG(4, "HandleComms");
  s = CommsGet(&c, ev);
  if (!s)
    EDBUG_RETURN_;
  unknown = 0;
  s1 = NULL, s2 = NULL;
  word(s, 1, w);
  if (!strcmp(w, "set"))
    {
      word(s, 2, w);
      if (!strcmp(w, "clientname"))
	{
	  if (c->clientname)
	    Efree(c->clientname);
	  c->clientname = duplicate(atword(s, 3));
	}
      else if (!strcmp(w, "version"))
	{
	  if (c->version)
	    Efree(c->version);
	  c->version = duplicate(atword(s, 3));
	}
      else if (!strcmp(w, "author"))
	{
	  if (c->author)
	    Efree(c->author);
	  c->author = duplicate(atword(s, 3));
	}
      else if (!strcmp(w, "email"))
	{
	  if (c->email)
	    Efree(c->email);
	  c->email = duplicate(atword(s, 3));
	}
      else if (!strcmp(w, "web"))
	{
	  if (c->web)
	    Efree(c->web);
	  c->web = duplicate(atword(s, 3));
	}
      else if (!strcmp(w, "address"))
	{
	  if (c->address)
	    Efree(c->address);
	  c->address = duplicate(atword(s, 3));
	}
      else if (!strcmp(w, "info"))
	{
	  if (c->info)
	    Efree(c->info);
	  c->info = duplicate(atword(s, 3));
	}
      else if (!strcmp(w, "pixmap"))
	{
	  word(s, 3, w);
	  sscanf(w, "%x", (int *)&c->pmap);
	}
      else
	unknown = 1;
    }
  else if (!strcmp(w, "reply"))
    {
      word(s, 2, w);
      if (!strcmp(w, "imageclass"))
	{
	  /* Reply format "reply imageclass NAME 24243" */
	  word(s, 3, w);
	  word(s, 4, w2);
	  HonorIclass(w, atoi(w2));
	}
      else
	unknown = 1;
    }
  else if (!strcmp(w, "restart"))
    {
      word(s, 2, w);
      if (w[0])
	{
	  char                sss[FILEPATH_LEN_MAX];

	  Esnprintf(sss, sizeof(sss), "restart_theme %s", w);
	  doExit(sss);
	}
      else
	doExit("restart");
    }
  else if (!strcmp(w, "list_bg"))
    {
      Background        **bg;
      int                 i, num, len = 0;
      char               *buf = NULL;

      bg = (Background **) ListItemType(&num, LIST_TYPE_BACKGROUND);
      if (bg)
	{
	  for (i = 0; i < num; i++)
	    {
	      len += strlen(bg[i]->name) + 1;
	      if (buf)
		buf = Erealloc(buf, len + 1);
	      else
		{
		  buf = Erealloc(buf, len + 1);
		  buf[0] = 0;
		}
	      strcat(buf, bg[i]->name);
	      strcat(buf, "\n");
	    }
	  Efree(bg);
	}
      if (buf)
	{
	  CommsSend(c, buf);
	  Efree(buf);
	}
      else
	CommsSend(c, "");
    }
  else if (!strcmp(w, "del_bg"))
    {
      Background         *bg;
      int                 i;
      char                dodel = 1;

      sscanf(s, "%*s %1000s", w);
      bg = (Background *) FindItem(w, 0, LIST_FINDBY_NAME, LIST_TYPE_BACKGROUND);
      if (bg)
	{
	  /* check for desktops referencing this bg - if there are - don't */
	  /* delete it */
	  for (i = 0; i < ENLIGHTENMENT_CONF_NUM_DESKTOPS; i++)
	    {
	      if (desks.desk[i].bg == bg)
		dodel = 0;
	    }
	  if (dodel)
	    {
	      bg = (Background *) RemoveItem(w, 0, LIST_FINDBY_NAME, LIST_TYPE_BACKGROUND);
	      if (bg->name)
		Efree(bg->name);
	      if (bg->bg.file)
		Efree(bg->bg.file);
	      if (bg->top.file)
		Efree(bg->top.file);
	      if (bg->pmap)
		Imlib_free_pixmap(id, bg->pmap);
	      Efree(bg);
	    }
	}
    }
  else if (!strcmp(w, "use_bg"))
    {
      Background         *bg;
      int                 i, wd;
      char                view;

      sscanf(s, "%*s %1000s", w);
      bg = (Background *) FindItem(w, 0, LIST_FINDBY_NAME, LIST_TYPE_BACKGROUND);
      if (bg)
	{
	  wd = 3;
	  w[0] = ' ';
	  while (w[0])
	    {
	      w[0] = 0;
	      word(s, wd++, w);
	      if (w[0])
		{
		  i = atoi(w);
		  if ((i >= 0) && (i < ENLIGHTENMENT_CONF_NUM_DESKTOPS))
		    {
		      if (desks.desk[i].bg)
			desks.desk[i].bg->last_viewed = 0;
		      view = desks.desk[i].viewable;
		      desks.desk[i].viewable = 0;
		      DesktopAccounting();
		      desks.desk[i].viewable = view;
		      desks.desk[i].bg = bg;
		      if (i < mode.numdesktops)
			{
			  if (desks.desk[i].viewable)
			    RefreshDesktop(i);
			}
		    }
		}
	    }
	}
    }
  else if (!strcmp(w, "uses_bg"))
    {
      Background         *bg;
      int                 i;
      char                buf[FILEPATH_LEN_MAX], buf2[FILEPATH_LEN_MAX];

      sscanf(s, "%*s %1000s", w);
      bg = (Background *) FindItem(w, 0, LIST_FINDBY_NAME, LIST_TYPE_BACKGROUND);
      buf[0] = 0;
      if (bg)
	{
	  for (i = 0; i < ENLIGHTENMENT_CONF_NUM_DESKTOPS; i++)
	    {
	      if (desks.desk[i].bg == bg)
		{
		  Esnprintf(buf2, sizeof(buf2), "%i\n", i);
		  strcat(buf, buf2);
		}
	    }
	}
      CommsSend(c, buf);
    }
  else if (!strcmp(w, "get_bg"))
    {
      Background         *bg;
      char                buf[FILEPATH_LEN_MAX];

      sscanf(s, "%*s %1000s", w);
      bg = (Background *) FindItem(w, 0, LIST_FINDBY_NAME, LIST_TYPE_BACKGROUND);
      Esnprintf(buf, sizeof(buf), "(null)");
      if (bg)
	{
	  if ((bg->bg.file) && (bg->top.file))
	    Esnprintf(buf, sizeof(buf),
		      "%s %i %i %i %s %i %i %i %i %i %i %s %i %i %i %i %i",
		      bg->name,
		      bg->bg.solid.r, bg->bg.solid.g, bg->bg.solid.b,
		      bg->bg.file, bg->bg.tile, bg->bg.keep_aspect,
		      bg->bg.xjust, bg->bg.yjust, bg->bg.xperc, bg->bg.yperc,
		      bg->top.file, bg->top.keep_aspect, bg->top.xjust,
		      bg->top.yjust, bg->top.xperc, bg->top.yperc);
	  else if ((!(bg->bg.file)) && (bg->top.file))
	    Esnprintf(buf, sizeof(buf),
		      "%s %i %i %i %s %i %i %i %i %i %i %s %i %i %i %i %i",
		      bg->name,
		      bg->bg.solid.r, bg->bg.solid.g, bg->bg.solid.b,
		      "(null)", bg->bg.tile, bg->bg.keep_aspect,
		      bg->bg.xjust, bg->bg.yjust, bg->bg.xperc, bg->bg.yperc,
		      bg->top.file, bg->top.keep_aspect, bg->top.xjust,
		      bg->top.yjust, bg->top.xperc, bg->top.yperc);
	  else if ((bg->bg.file) && (!(bg->top.file)))
	    Esnprintf(buf, sizeof(buf),
		      "%s %i %i %i %s %i %i %i %i %i %i %s %i %i %i %i %i",
		      bg->name,
		      bg->bg.solid.r, bg->bg.solid.g, bg->bg.solid.b,
		      bg->bg.file, bg->bg.tile, bg->bg.keep_aspect,
		      bg->bg.xjust, bg->bg.yjust, bg->bg.xperc, bg->bg.yperc,
		      "(null)", bg->top.keep_aspect, bg->top.xjust,
		      bg->top.yjust, bg->top.xperc, bg->top.yperc);
	  else if ((!(bg->bg.file)) && (!(bg->top.file)))
	    Esnprintf(buf, sizeof(buf),
		      "%s %i %i %i %s %i %i %i %i %i %i %s %i %i %i %i %i",
		      bg->bg.solid.r, bg->bg.solid.g, bg->bg.solid.b,
		      "(null)", bg->bg.tile, bg->bg.keep_aspect,
		      bg->bg.xjust, bg->bg.yjust, bg->bg.xperc, bg->bg.yperc,
		      "(null)", bg->top.keep_aspect, bg->top.xjust,
		      bg->top.yjust, bg->top.xperc, bg->top.yperc);
	}
      CommsSend(c, buf);
    }
  else if (!strcmp(w, "set_bg"))
    {
      Background         *bg;
      ImlibColor          icl;
      int                 i;
      char                tmp[1024];
      char               *name = NULL, *bgf = NULL, *topf = NULL;
      char                updated = 0, tile, keep_aspect, tkeep_aspect;
      int                 xjust, yjust, xperc, yperc, txjust, tyjust, txperc,
                          typerc;

      sscanf(s, "%*s %1000s", w);
      bg = (Background *) FindItem(w, 0, LIST_FINDBY_NAME, LIST_TYPE_BACKGROUND);
      icl.r = 99;
      i = sscanf(s, "%1000s %1000s %i %i %i %1000s %i %i %i %i %i %i %1000s %i %i %i %i %i",
		 tmp, tmp,
	 &(icl.r), &(icl.g), &(icl.b), tmp, (int *)&tile, (int *)&keep_aspect,
	   &xjust, &yjust, &xperc, &yperc, tmp, (int *)&tkeep_aspect, &txjust,
		 &tyjust, &txperc, &typerc);
      if (bg)
	{
	  name = duplicate(w);
	  word(s, 6, w);
	  if (strcmp("(null)", w))
	    bgf = duplicate(w);
	  word(s, 13, w);
	  if (strcmp("(null)", w))
	    topf = duplicate(w);
	  word(s, 13, w);

	  if (icl.r != bg->bg.solid.r)
	    updated = 1;
	  if (icl.g != bg->bg.solid.g)
	    updated = 1;
	  if (icl.b != bg->bg.solid.b)
	    updated = 1;
	  bg->bg.solid.r = icl.r;
	  bg->bg.solid.g = icl.g;
	  bg->bg.solid.b = icl.b;
	  if ((!bg->bg.file) || (!bgf))
	    {
	      if ((bg->bg.file) && (bgf))
		{
		  if (strcmp(bg->bg.file, bgf))
		    updated = 1;
		}
	      else
		updated = 1;
	      if (bg->bg.file)
		Efree(bg->bg.file);
	      bg->bg.file = bgf;
	      if (tile != bg->bg.tile)
		updated = 1;
	      if (keep_aspect != bg->bg.keep_aspect)
		updated = 1;
	      if (xjust != bg->bg.xjust)
		updated = 1;
	      if (yjust != bg->bg.yjust)
		updated = 1;
	      if (xperc != bg->bg.xperc)
		updated = 1;
	      if (yperc != bg->bg.yperc)
		updated = 1;
	      bg->bg.tile = tile;
	      bg->bg.keep_aspect = keep_aspect;
	      bg->bg.xjust = xjust;
	      bg->bg.yjust = yjust;
	      bg->bg.xperc = xperc;
	      bg->bg.yperc = yperc;
	    }
	  if ((!bg->top.file) || (!topf))
	    {
	      if ((bg->top.file) && (topf))
		{
		  if (strcmp(bg->top.file, topf))
		    updated = 1;
		}
	      else
		updated = 1;
	      if (bg->top.file)
		Efree(bg->top.file);
	      bg->top.file = topf;
	      if (tkeep_aspect != bg->top.keep_aspect)
		updated = 1;
	      if (txjust != bg->top.xjust)
		updated = 1;
	      if (tyjust != bg->top.yjust)
		updated = 1;
	      if (txperc != bg->top.xperc)
		updated = 1;
	      if (typerc != bg->top.yperc)
		updated = 1;
	      bg->top.keep_aspect = tkeep_aspect;
	      bg->top.xjust = txjust;
	      bg->top.yjust = tyjust;
	      bg->top.xperc = txperc;
	      bg->top.yperc = typerc;
	    }
	  if (updated)
	    {
	      if (bg->pmap)
		Imlib_free_pixmap(id, bg->pmap);
	      bg->pmap = 0;
	      for (i = 0; i < ENLIGHTENMENT_CONF_NUM_DESKTOPS; i++)
		{
		  if ((desks.desk[i].bg == bg) &&
		      (desks.desk[i].viewable))
		    RefreshDesktop(i);
		}
	    }
	}
      else
	{
	  name = duplicate(w);
	  word(s, 6, w);
	  if (strcmp("(null)", w))
	    bgf = duplicate(w);
	  word(s, 13, w);
	  if (strcmp("(null)", w))
	    topf = duplicate(w);
	  word(s, 13, w);
	  bg = CreateDesktopBG(name, &icl, bgf, tile, keep_aspect, xjust,
			       yjust, xperc, yperc, topf, tkeep_aspect,
			       txjust, tyjust, txperc, typerc);
	  if (name)
	    Efree(name);
	  if (bgf)
	    Efree(bgf);
	  if (topf)
	    Efree(topf);
	  AddItem(bg, bg->name, 0, LIST_TYPE_BACKGROUND);
	}
    }
  else if (!strcmp(w, "draw_bg_to"))
    {
      Window              win;
      Background         *bg;

      sscanf(s, "%*s %x %1000s", (unsigned int *)&win, w);
      bg = (Background *) FindItem(w, 0, LIST_FINDBY_NAME, LIST_TYPE_BACKGROUND);
      if (bg)
	SetBackgroundTo(id, win, bg, 0);
      CommsSend(c, "done");
    }
  else if (!strcmp(w, "set_controls"))
    {
      int                 i, num, wd;
      EWin              **lst;
      Button            **blst;
      int                 a, b;

      wd = 2;
      w[0] = 0;
      while (atword(s, wd))
	{
	  word(s, wd, w);
	  wd++;
	  if (!strcmp(w, "FOCUSMODE:"))
	    {
	      word(s, wd, w);
	      mode.focusmode = atoi(w);
	    }
	  else if (!strcmp(w, "DOCKDIRMODE:"))
	    {
	      word(s, wd, w);
	      mode.dockdirmode = atoi(w);
	    }
	  else if (!strcmp(w, "ICONDIRMODE:"))
	    {
	      word(s, wd, w);
	      mode.primaryicondir = atoi(w);
	    }
	  else if (!strcmp(w, "MOVEMODE:"))
	    {
	      word(s, wd, w);
	      mode.movemode = atoi(w);
	      if ((ird) && (mode.movemode == 5))
		mode.movemode = 3;
	    }
	  else if (!strcmp(w, "RESIZEMODE:"))
	    {
	      word(s, wd, w);
	      mode.resizemode = atoi(w);
	      if (mode.resizemode == 5)
		mode.resizemode = 3;
	    }
	  else if (!strcmp(w, "SLIDEMODE:"))
	    {
	      word(s, wd, w);
	      mode.slidemode = atoi(w);
	    }
	  else if (!strcmp(w, "CLEANUPSLIDE:"))
	    {
	      word(s, wd, w);
	      mode.cleanupslide = atoi(w);
	    }
	  else if (!strcmp(w, "MAPSLIDE:"))
	    {
	      word(s, wd, w);
	      mode.mapslide = atoi(w);
	    }
	  else if (!strcmp(w, "SLIDESPEEDMAP:"))
	    {
	      word(s, wd, w);
	      mode.slidespeedmap = atoi(w);
	    }
	  else if (!strcmp(w, "SLIDESPEEDCLEANUP:"))
	    {
	      word(s, wd, w);
	      mode.slidespeedcleanup = atoi(w);
	    }
	  else if (!strcmp(w, "SHADESPEED:"))
	    {
	      word(s, wd, w);
	      mode.shadespeed = atoi(w);
	    }
	  else if (!strcmp(w, "DESKTOPBGTIMEOUT:"))
	    {
	      word(s, wd, w);
	      mode.desktop_bg_timeout = atoi(w);
	    }
	  else if (!strcmp(w, "SOUND:"))
	    {
	      word(s, wd, w);
	      mode.sound = atoi(w);
	    }
	  else if (!strcmp(w, "BUTTONMOVERESISTANCE:"))
	    {
	      word(s, wd, w);
	      mode.button_move_resistance = atoi(w);
	    }
	  else if (!strcmp(w, "AUTOSAVE:"))
	    {
	      word(s, wd, w);
	      mode.autosave = atoi(w);
	    }
	  else if (!strcmp(w, "MEMORYPARANOIA:"))
	    {
	      word(s, wd, w);
	      mode.memory_paranoia = atoi(w);
	    }
	  else if (!strcmp(w, "MENUSLIDE:"))
	    {
	      word(s, wd, w);
	      mode.menuslide = atoi(w);
	    }
	  else if (!strcmp(w, "NUMDESKTOPS:"))
	    {
	      word(s, wd, w);
	      mode.numdesktops = atoi(w);
	      if (mode.numdesktops <= 0)
		mode.numdesktops = 1;
	      else if (mode.numdesktops > ENLIGHTENMENT_CONF_NUM_DESKTOPS)
		mode.numdesktops = ENLIGHTENMENT_CONF_NUM_DESKTOPS;
	    }
	  else if (!strcmp(w, "TOOLTIPS:"))
	    {
	      word(s, wd, w);
	      mode.tooltips = atoi(w);
	    }
	  else if (!strcmp(w, "TIPTIME:"))
	    {
	      word(s, wd, w);
	      mode.tiptime = atof(w);
	    }
	  else if (!strcmp(w, "AUTORAISE:"))
	    {
	      word(s, wd, w);
	      mode.autoraise = atoi(w);
	    }
	  else if (!strcmp(w, "AUTORAISETIME:"))
	    {
	      word(s, wd, w);
	      mode.autoraisetime = atof(w);
	    }
	  else if (!strcmp(w, "DOCKSTARTX:"))
	    {
	      word(s, wd, w);
	      mode.dockstartx = atoi(w);
	    }
	  else if (!strcmp(w, "DOCKSTARTY:"))
	    {
	      word(s, wd, w);
	      mode.dockstarty = atoi(w);
	    }
	  else if (!strcmp(w, "SAVEUNDER:"))
	    {
	      word(s, wd, w);
	      mode.save_under = atoi(w);
	    }
	  else if (!strcmp(w, "DRAGDIR:"))
	    {
	      word(s, wd, w);
	      desks.dragdir = atoi(w);
	    }
	  else if (!strcmp(w, "DRAGBARWIDTH:"))
	    {
	      word(s, wd, w);
	      desks.dragbar_width = atoi(w);
	    }
	  else if (!strcmp(w, "DRAGBARORDERING:"))
	    {
	      word(s, wd, w);
	      desks.dragbar_ordering = atoi(w);
	    }
	  else if (!strcmp(w, "DRAGBARLENGTH:"))
	    {
	      word(s, wd, w);
	      desks.dragbar_length = atoi(w);
	    }
	  else if (!strcmp(w, "DESKSLIDEIN:"))
	    {
	      word(s, wd, w);
	      desks.slidein = atoi(w);
	    }
	  else if (!strcmp(w, "DESKSLIDESPEED:"))
	    {
	      word(s, wd, w);
	      desks.slidespeed = atoi(w);
	    }
	  else if (!strcmp(w, "HIQUALITYBG:"))
	    {
	      word(s, wd, w);
	      desks.hiqualitybg = atoi(w);
	    }
	  else if (!strcmp(w, "TRANSIENTSFOLLOWLEADER:"))
	    {
	      word(s, wd, w);
	      mode.transientsfollowleader = atoi(w);
	    }
	  else if (!strcmp(w, "SWITCHFORTRANSIENTMAP:"))
	    {
	      word(s, wd, w);
	      mode.switchfortransientmap = atoi(w);
	    }
	  else if (!strcmp(w, "SHOWICONS:"))
	    {
	      word(s, wd, w);
	      mode.showicons = atoi(w);
	    }
	  else if (!strcmp(w, "SHOWICONS:"))
	    {
	      word(s, wd, w);
	      mode.showicons = atoi(w);
	    }
	  else if (!strcmp(w, "AREA_SIZE:"))
	    {
	      w[0] = 0;
	      word(s, wd, w);
	      if (w[0])
		a = atoi(w);
	      else
		a = 0;
	      wd++;
	      w[0] = 0;
	      word(s, wd, w);
	      if (w[0])
		b = atoi(w);
	      else
		b = 0;
	      if ((a > 0) && (b > 0))
		SetAreaSize(a, b);
	    }
	  else
	    {
	      Alert("unknown variable %500s\n", w);
	    }
	  wd++;
	}
      FixFocus();
      GNOME_SetDeskCount();
      GNOME_SetDeskNames();
      lst = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
      if (lst)
	{
	  for (i = 0; i < num; i++)
	    {
	      if (lst[i]->desktop >= mode.numdesktops)
		MoveEwinToDesktop(lst[i], mode.numdesktops - 1);
	    }
	  Efree(lst);
	}
      blst = (Button **) ListItemType(&num, LIST_TYPE_BUTTON);
      if (lst)
	{
	  for (i = 0; i < num; i++)
	    {
	      if (!strcmp(blst[i]->name, "ICON"))
		{
		  if (mode.showicons)
		    ShowButton(blst[i]);
		  else
		    HideButton(blst[i]);
		}
	    }
	  Efree(blst);
	}
      for (i = mode.numdesktops; i < ENLIGHTENMENT_CONF_NUM_DESKTOPS; i++)
	LowerDesktop(i);
    }
  else if (!strcmp(w, "get_controls"))
    {
      char                buf[FILEPATH_LEN_MAX];
      int                 a, b;

      GetAreaSize(&a, &b);
      Esnprintf(buf, sizeof(buf),
		"FOCUSMODE: %i\n"
		"DOCKDIRMODE: %i\n"
		"ICONDIRMODE: %i\n"
		"MOVEMODE: %i\n"
		"RESIZEMODE: %i\n"
		"SLIDEMODE: %i\n"
		"CLEANUPSLIDE: %i\n"
		"MAPSLIDE: %i\n"
		"SLIDESPEEDMAP: %i\n"
		"SLIDESPEEDCLEANUP: %i\n"
		"SHADESPEED: %i\n"
		"DESKTOPBGTIMEOUT: %i\n"
		"SOUND: %i\n"
		"BUTTONMOVERESISTANCE: %i\n"
		"AUTOSAVE: %i\n"
		"MEMORYPARANOIA: %i\n"
		"TOOLTIPS: %i\n"
		"TIPTIME: %f\n"
		"AUTORAISE: %i\n"
		"AUTORAISETIME: %f\n"
		"DOCKSTARTX: %i\n"
		"DOCKSTARTY: %i\n"
		"SAVEUNDER: %i\n"
		"MENUSLIDE: %i\n"
		"NUMDESKTOPS: %i\n"

		"DRAGDIR: %i\n"
		"DRAGBARWIDTH: %i\n"
		"DRAGBARORDERING: %i\n"
		"DRAGBARLENGTH: %i\n"
		"DESKSLIDEIN: %i\n"
		"DESKSLIDESPEED: %i\n"
		"HIQUALITYBG: %i\n"

		"TRANSIENTSFOLLOWLEADER: %i\n"
		"SWITCHFORTRANSIENTMAP: %i\n"
		"SHOWICONS: %i\n"
		"AREA_SIZE: %i %i\n"
		,
		mode.focusmode, mode.dockdirmode, mode.primaryicondir,
		mode.movemode, mode.resizemode, mode.slidemode,
		mode.cleanupslide, mode.mapslide, mode.slidespeedmap,
		mode.slidespeedcleanup, mode.shadespeed,
		mode.desktop_bg_timeout, mode.sound,
		mode.button_move_resistance, mode.autosave,
		mode.memory_paranoia, mode.tooltips, mode.tiptime,
		mode.autoraise, mode.autoraisetime,
		mode.dockstartx, mode.dockstarty, mode.save_under,
		mode.menuslide, mode.numdesktops,

		desks.dragdir, desks.dragbar_width,
		desks.dragbar_ordering, desks.dragbar_length, desks.slidein,
		desks.slidespeed, desks.hiqualitybg,

		mode.transientsfollowleader, mode.switchfortransientmap,
		mode.showicons, a, b
	);
      CommsSend(c, buf);
    }
  else if (!strcmp(w, "call_raw"))
    {
      char               *par;
      int                 aid;

      word(s, 2, w);
      aid = atoi(w);
      par = atword(s, 3);
      if ((aid > 0) && (aid < ACTION_NUMBEROF))
	(*(ActionFunctions[aid])) (par);
    }
  else if (!strcmp(w, "goto_desktop"))
    {
      word(s, 2, w);
      if (w[0])
	{
	  GotoDesktop(atoi(w));
	}
    }
  else if (!strcmp(w, "list_clients"))
    {
      char                buf[FILEPATH_LEN_MAX], *ret = NULL, none[] = "-NONE-";
      EWin              **lst;
      int                 i, num;

      lst = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
      for (i = 0; i < num; i++)
	{
	  if (!lst[i]->client.title)
	    lst[i]->client.title = none;
	  if (!lst[i]->client.class)
	    lst[i]->client.class = none;
	  if (!lst[i]->client.name)
	    lst[i]->client.name = none;
	  if (!lst[i]->client.command)
	    lst[i]->client.command = none;
	  if (!lst[i]->client.machine)
	    lst[i]->client.machine = none;
	  if (!lst[i]->client.icon_name)
	    lst[i]->client.icon_name = none;
	  if (lst[i]->border)
	    {
	      Esnprintf(buf, sizeof(buf),
			"***CLIENT***\n"
			"CLIENT_WIN_ID:          %8x\n"
			"FRAME_WIN_ID:           %8x\n"
			"FRAME_X,Y:              %5i , %5i\n"
			"FRAME_WIDTH,HEIGHT:     %5i , %5i\n"
			"BORDER_NAME:            %s\n"
			"BORDER_BORDER:          %5i , %5i , %5i , %5i\n"
			"DESKTOP_NUMBER:         %5i\n"
			"INTERNAL_GROUP_ID:      %5i\n"
			"DOCKED:                 %5i\n"
			"STICKY:                 %5i\n"
			"VISIBLE:                %5i\n"
			"ICONIFIED:              %5i\n"
			"SHADED:                 %5i\n"
			"ACTIVE:                 %5i\n"
			"LAYER:                  %5i\n"
			"NEVER_USE_AREA:         %5i\n"
			"FLOATING:               %5i\n"
			"CLIENT_WIDTH,HEIGHT:    %5i , %5i\n"
			"ICON_WIN_ID:            %8x\n"
			"ICON_PIXMAP,MASK_ID:    %8x , %8x\n"
			"CLIENT_GROUP_LEADER_ID: %8x\n"
			"CLIENT_NEEDS_INPUT:     %5i\n"
			"TRANSIENT:              %5i\n"
			"TITLE:                  %s\n"
			"CLASS:                  %s\n"
			"NAME:                   %s\n"
			"COMMAND:                %s\n"
			"MACHINE:                %s\n"
			"ICON_NAME:              %s\n"
			"IS_GROUP_LEADER:        %5i\n"
			"NO_RESIZE_HORIZONTAL:   %5i\n"
			"NO_RESIZE_VERTICAL:     %5i\n"
			"SHAPED:                 %5i\n"
			"MIN_WIDTH,HEIGHT:       %5i , %5i\n"
			"MAX_WIDTH,HEIGHT:       %5i , %5i\n"
			"BASE_WIDTH,HEIGHT:      %5i , %5i\n"
			"WIDTH,HEIGHT_INC:       %5i , %5i\n"
			"ASPECT_MIN,MAX:         %5.5f , %5.5f\n"
			"MWM_BORDER:             %5i\n"
			"MWM_RESIZEH:            %5i\n"
			"MWM_TITLE:              %5i\n"
			"MWM_MENU:               %5i\n"
			"MWM_MINIMIZE:           %5i\n"
			"MWM_MAXIMIZE:           %5i\n"
			"APP_STATE:              %5i\n",
			lst[i]->client.win,
			lst[i]->win,
			lst[i]->x,
			lst[i]->y,
			lst[i]->w,
			lst[i]->h,
			lst[i]->border->name,
			lst[i]->border->border.left,
			lst[i]->border->border.right,
			lst[i]->border->border.top,
			lst[i]->border->border.bottom,
			lst[i]->desktop,
			lst[i]->group,
			lst[i]->docked,
			lst[i]->sticky,
			lst[i]->visible,
			lst[i]->iconified,
			lst[i]->shaded,
			lst[i]->active,
			lst[i]->layer,
			lst[i]->never_use_area,
			lst[i]->floating,
			lst[i]->client.w,
			lst[i]->client.h,
			lst[i]->client.icon_win,
			lst[i]->client.icon_pmap,
			lst[i]->client.icon_mask,
			lst[i]->client.group,
			lst[i]->client.need_input,
			lst[i]->client.transient,
			lst[i]->client.title,
			lst[i]->client.class,
			lst[i]->client.name,
			lst[i]->client.command,
			lst[i]->client.machine,
			lst[i]->client.icon_name,
			lst[i]->client.is_group_leader,
			lst[i]->client.no_resize_h,
			lst[i]->client.no_resize_v,
			lst[i]->client.shaped,
			lst[i]->client.width.min,
			lst[i]->client.height.min,
			lst[i]->client.width.max,
			lst[i]->client.height.max,
			lst[i]->client.base_w,
			lst[i]->client.base_h,
			lst[i]->client.w_inc,
			lst[i]->client.h_inc,
			lst[i]->client.aspect_min,
			lst[i]->client.aspect_max,
			lst[i]->client.mwm_decor_border,
			lst[i]->client.mwm_decor_resizeh,
			lst[i]->client.mwm_decor_title,
			lst[i]->client.mwm_decor_menu,
			lst[i]->client.mwm_decor_minimize,
			lst[i]->client.mwm_decor_maximize,
			lst[i]->client.app_state);
	    }
	  else
	    {
	      Esnprintf(buf, sizeof(buf),
			"***CLIENT***\n"
			"CLIENT_WIN_ID:          %8x\n"
			"FRAME_WIN_ID:           %8x\n"
			"FRAME_X,Y:              %5i , %5i\n"
			"FRAME_WIDTH,HEIGHT:     %5i , %5i\n"
			"BORDER_NAME:            %s\n"
			"BORDER_BORDER:          %5i , %5i , %5i , %5i\n"
			"DESKTOP_NUMBER:         %5i\n"
			"INTERNAL_GROUP_ID:      %5i\n"
			"DOCKED:                 %5i\n"
			"STICKY:                 %5i\n"
			"VISIBLE:                %5i\n"
			"ICONIFIED:              %5i\n"
			"SHADED:                 %5i\n"
			"ACTIVE:                 %5i\n"
			"LAYER:                  %5i\n"
			"NEVER_USE_AREA:         %5i\n"
			"FLOATING:               %5i\n"
			"CLIENT_WIDTH,HEIGHT:    %5i , %5i\n"
			"ICON_WIN_ID:            %8x\n"
			"ICON_PIXMAP,MASK_ID:    %8x , %8x\n"
			"CLIENT_GROUP_LEADER_ID: %8x\n"
			"CLIENT_NEEDS_INPUT:     %5i\n"
			"TRANSIENT:              %5i\n"
			"TITLE:                  %s\n"
			"CLASS:                  %s\n"
			"NAME:                   %s\n"
			"COMMAND:                %s\n"
			"MACHINE:                %s\n"
			"ICON_NAME:              %s\n"
			"IS_GROUP_LEADER:        %5i\n"
			"NO_RESIZE_HORIZONTAL:   %5i\n"
			"NO_RESIZE_VERTICAL:     %5i\n"
			"SHAPED:                 %5i\n"
			"MIN_WIDTH,HEIGHT:       %5i , %5i\n"
			"MAX_WIDTH,HEIGHT:       %5i , %5i\n"
			"BASE_WIDTH,HEIGHT:      %5i , %5i\n"
			"WIDTH,HEIGHT_INC:       %5i , %5i\n"
			"ASPECT_MIN,MAX:         %5.5f , %5.5f\n"
			"MWM_BORDER:             %5i\n"
			"MWM_RESIZEH:            %5i\n"
			"MWM_TITLE:              %5i\n"
			"MWM_MENU:               %5i\n"
			"MWM_MINIMIZE:           %5i\n"
			"MWM_MAXIMIZE:           %5i\n"
			"APP_STATE:              %5i\n",
			lst[i]->client.win,
			lst[i]->win,
			lst[i]->x,
			lst[i]->y,
			lst[i]->w,
			lst[i]->h,
			none,
			0,
			0,
			0,
			0,
			lst[i]->desktop,
			lst[i]->group,
			lst[i]->docked,
			lst[i]->sticky,
			lst[i]->visible,
			lst[i]->iconified,
			lst[i]->shaded,
			lst[i]->active,
			lst[i]->layer,
			lst[i]->never_use_area,
			lst[i]->floating,
			lst[i]->client.w,
			lst[i]->client.h,
			lst[i]->client.icon_win,
			lst[i]->client.icon_pmap,
			lst[i]->client.icon_mask,
			lst[i]->client.group,
			lst[i]->client.need_input,
			lst[i]->client.transient,
			lst[i]->client.title,
			lst[i]->client.class,
			lst[i]->client.name,
			lst[i]->client.command,
			lst[i]->client.machine,
			lst[i]->client.icon_name,
			lst[i]->client.is_group_leader,
			lst[i]->client.no_resize_h,
			lst[i]->client.no_resize_v,
			lst[i]->client.shaped,
			lst[i]->client.width.min,
			lst[i]->client.height.min,
			lst[i]->client.width.max,
			lst[i]->client.height.max,
			lst[i]->client.base_w,
			lst[i]->client.base_h,
			lst[i]->client.w_inc,
			lst[i]->client.h_inc,
			lst[i]->client.aspect_min,
			lst[i]->client.aspect_max,
			lst[i]->client.mwm_decor_border,
			lst[i]->client.mwm_decor_resizeh,
			lst[i]->client.mwm_decor_title,
			lst[i]->client.mwm_decor_menu,
			lst[i]->client.mwm_decor_minimize,
			lst[i]->client.mwm_decor_maximize,
			lst[i]->client.app_state);
	    }
	  if (lst[i]->client.title == none)
	    lst[i]->client.title = NULL;
	  if (lst[i]->client.class == none)
	    lst[i]->client.class = NULL;
	  if (lst[i]->client.name == none)
	    lst[i]->client.name = NULL;
	  if (lst[i]->client.command == none)
	    lst[i]->client.command = NULL;
	  if (lst[i]->client.machine == none)
	    lst[i]->client.machine = NULL;
	  if (lst[i]->client.icon_name == none)
	    lst[i]->client.icon_name = NULL;
	  if (!ret)
	    {
	      ret = Emalloc(strlen(buf) + 1);
	      ret[0] = 0;
	    }
	  else
	    {
	      ret = Erealloc(ret, strlen(ret) + strlen(buf) + 1);
	    }
	  strcat(ret, buf);
	}
      if (ret)
	{
	  CommsSend(c, ret);
	  Efree(ret);
	}
      if (lst)
	Efree(lst);
    }
  else
    unknown = 1;

  if (unknown)
    {
      s1 = c->clientname;
      s2 = c->version;
      if (!s1)
	s1 = sunknown;
      if (!s2)
	s2 = sunknown;
      Alert("Received Unknown Client Message.\n"
	    "Client Name:    %s\n"
	    "Client Version: %s\n"
	    "Message Contents:\n\n"
	    "%s\n", s1, s2, s);
    }
  Efree(s);
  EDBUG_RETURN_;
}

void
DisplayClientInfo(Client * c, int onoff)
{
  EDBUG(6, "DisplayClientInfo");
  if (!c)
    EDBUG_RETURN_;

  printf("NAME ID:\n%s\n", c->name);
  printf("CLIENT COMMS WINDOW:\n%i\n", (int)c->win);
  if (onoff)
    {
      if (c->clientname)
	{
	  printf("CLIENT NAME:\n%s\n", c->clientname);
	  if (c->version)
	    {
	      printf("VERSION:\n%s\n", c->version);
	    }
	  if (c->email)
	    {
	      printf("CONTACT EMAIL:\n%s\n", c->email);
	    }
	  if (c->web)
	    {
	      printf("CLIENT WEB PAGE:\n%s\n", c->web);
	    }
	  if (c->address)
	    {
	      printf("AUTHOR ADDRESS:\n%s\n", c->address);
	    }
	  if (c->info)
	    {
	      printf("CLIENT INFORMATION:\n%s\n", c->info);
	    }
	}
    }
  else
    {
    }
  EDBUG_RETURN_;
}

void
HideClientInfo()
{
  EDBUG(6, "HideClientInfo");
  EDBUG_RETURN_;
}
