#include "E.h"

void
HandleDrawQueue()
{
  DrawQueue         **lst = NULL, *dq;
  int                 i, num;
  char                already, p_queue;

  EDBUG(4, "HandleDrawQueue");
  if ((mode.mode == MODE_MOVE) && (mode.movemode > 0))
    EDBUG_RETURN_;
  if ((mode.mode == MODE_RESIZE) && (mode.resizemode > 0))
    EDBUG_RETURN_;
  if ((mode.mode == MODE_RESIZE_H) && (mode.resizemode > 0))
    EDBUG_RETURN_;
  if ((mode.mode == MODE_RESIZE_V) && (mode.resizemode > 0))
    EDBUG_RETURN_;
  p_queue = queue_up;
  queue_up = 0;
  num = 0;
  /* find all DRAW queue entries most recent first and add them to the */
  /* end of the draw list array if there are no previous entries for that */
  /* draw type and that window in the array */
  while ((dq = (DrawQueue *) RemoveItem(NULL, 0, LIST_FINDBY_NONE,
					LIST_TYPE_DRAW)))
    {
      already = 0;
      if (dq->shape_propagate)
	{
	  for (i = 0; i < num; i++)
	    {
	      if ((lst[i]->win == dq->win) && (lst[i]->shape_propagate))
		{
		  already = 1;
		  i = num;
		}
	    }
	}
      else if (dq->text)
	{
	  for (i = 0; i < num; i++)
	    {
	      if ((lst[i]->win == dq->win) && (lst[i]->text))
		{
		  Efree(dq->text);
		  already = 1;
		  i = num;
		}
	    }
	}
      else
	{
	  for (i = 0; i < num; i++)
	    {
	      if ((lst[i]->win == dq->win) && (!lst[i]->shape_propagate)
		  && (!lst[i]->text))
		{
		  already = 1;
		  i = num;
		}
	    }
	}
      if (already)
	{
	  if (dq)
	    Efree(dq);
	}
      else
	{
	  num++;
	  lst = Erealloc(lst, num * sizeof(DrawQueue *));
	  lst[num - 1] = dq;
	}
    }
  /* go thru the list in chronological order (ie reverse) and do the draws */
  if (lst)
    {
      for (i = num - 1; i >= 0; i--)
	{
	  if (lst[i]->shape_propagate)
	    PropagateShapes(lst[i]->win);
	  else if (lst[i]->text)
	    {
	      TclassApply(lst[i]->iclass, lst[i]->win, lst[i]->w, lst[i]->h,
			  lst[i]->active, lst[i]->sticky, lst[i]->state,
			  lst[i]->expose, lst[i]->tclass, lst[i]->text);
	      Efree(lst[i]->text);
	    }
	  else
	    {
	      IclassApply(lst[i]->iclass, lst[i]->win, lst[i]->w, lst[i]->h,
			  lst[i]->active, lst[i]->sticky, lst[i]->state,
			  0);
	    }
	  Efree(lst[i]);
	}
      Efree(lst);
    }
  queue_up = p_queue;
  EDBUG_RETURN_;
}

Pixmap
EInitPixmap(Window win)
{
  Pixmap              pmap;

  pmap = XCreatePixmap(disp, win, root.w, root.h, root.depth);
  return pmap;
}

void
EFillPixmap(Window win, Pixmap pmap, int x, int y, int w, int h)
{
  XGCValues           gcv;
  GC                  gc;

  gcv.subwindow_mode = IncludeInferiors;
  gc = XCreateGC(disp, win, GCSubwindowMode, &gcv);
  XCopyArea(disp, win, pmap, gc, x, y, w, h, x, y);
  XFreeGC(disp, gc);
  return;
}

void
EPastePixmap(Window win, Pixmap pmap, int x, int y, int w, int h)
{
  XGCValues           gcv;
  GC                  gc;

  gcv.subwindow_mode = IncludeInferiors;
  gc = XCreateGC(disp, win, GCSubwindowMode, &gcv);
  XCopyArea(disp, pmap, win, gc, x, y, w, h, x, y);
  XFreeGC(disp, gc);
}

XImage             *
EInitImage(XShmSegmentInfo ** shminfo, EWin * ewin)
{
  int                 x, y, w, h;
  XImage             *xim;

  if ((ewin->sticky) || (ewin->floating))
    {
      x = ewin->x;
      y = ewin->y;
    }
  else
    {
      x = ewin->x + desks.desk[ewin->desktop].x;
      y = ewin->y + desks.desk[ewin->desktop].y;
    }
  w = ewin->w;
  h = ewin->h;

  if (x < 0)
    {
      w += x;
      x = 0;
    }
  if ((x + w) > root.w)
    w -= (x + w) - root.w;
  if (w < 0)
    w = 0;
  if (y < 0)
    {
      h += y;
      y = 0;
    }
  if ((y + h) > root.h)
    h -= (y + h) - root.h;
  if (h < 0)
    h = 0;
  if ((w > 0) && (h > 0))
    {
      if (id->x.shm)
	{
	  *shminfo = Emalloc(sizeof(XShmSegmentInfo));
	  xim = XShmCreateImage(disp, root.vis, root.depth, ZPixmap, NULL,
				(*shminfo), w, h);
	  (*shminfo)->shmid = shmget(IPC_PRIVATE, xim->bytes_per_line *
				     xim->height, IPC_CREAT | 0666);
	  (*shminfo)->shmaddr = xim->data = shmat((*shminfo)->shmid, 0, 0);
	  (*shminfo)->readOnly = False;
	  XShmAttach(disp, (*shminfo));
	  return xim;
	}
    }
  *shminfo = NULL;
  return NULL;
}

XImage             *
EFillImage(XImage * xim, XShmSegmentInfo ** shminfo, EWin * ewin, Drawable win)
{
  XImage             *xim2;
  int                 x, y, w, h;

  if ((ewin->sticky) || (ewin->floating))
    {
      x = ewin->x;
      y = ewin->y;
    }
  else
    {
      x = ewin->x + desks.desk[ewin->desktop].x;
      y = ewin->y + desks.desk[ewin->desktop].y;
    }
  w = ewin->w;
  h = ewin->h;

  if (x < 0)
    {
      w += x;
      x = 0;
    }
  if ((x + w) > root.w)
    w -= (x + w) - root.w;
  if (w < 0)
    w = 0;
  if (y < 0)
    {
      h += y;
      y = 0;
    }
  if ((y + h) > root.h)
    h -= (y + h) - root.h;
  if (h < 0)
    h = 0;
  if ((w > 0) && (h > 0))
    {
      if (*shminfo)
	{
	  XShmGetImage(disp, win, xim, x, y, 0xffffffff);
	  return xim;
	}
      if (xim)
	XDestroyImage(xim);
      xim2 = XGetImage(disp, win, x, y, w, h, 0xffffffff, ZPixmap);
      return xim2;
    }
  return NULL;
}

void
EDestroyImage(XImage * xim, XShmSegmentInfo ** shminfo)
{
  if (*shminfo)
    {
      XSync(disp, False);
      XShmDetach(disp, (*shminfo));
      shmdt((*shminfo)->shmaddr);
      shmctl((*shminfo)->shmid, IPC_RMID, 0);
      XDestroyImage(xim);
      Efree(*shminfo);
      return;
    }
  XDestroyImage(xim);
}

void
EBlendRemoveShape(EWin * ewin, Pixmap pmap)
{
  XGCValues           gcv;
  GC                  gc, gcm;
  int                 rn, ord;
  int                 i, x, y, w, h;
  XRectangle         *rl;
  Pixmap              mask;

  if ((ewin->sticky) || (ewin->floating))
    {
      x = ewin->x;
      y = ewin->y;
    }
  else
    {
      x = ewin->x + desks.desk[ewin->desktop].x;
      y = ewin->y + desks.desk[ewin->desktop].y;
    }
  w = ewin->w;
  h = ewin->h;

  if (x < 0)
    {
      w += x;
      x = 0;
    }
  if ((x + w) > root.w)
    w -= (x + w) - root.w;
  if (w < 0)
    w = 0;
  if (y < 0)
    {
      h += y;
      y = 0;
    }
  if ((y + h) > root.h)
    h -= (y + h) - root.h;
  if (h < 0)
    h = 0;
  if ((w <= 0) || (h <= 0))
    return;
  rl = NULL;
  rl = XShapeGetRectangles(disp, ewin->win, ShapeBounding, &rn, &ord);
  if (rn < 1)
    return;
  else if (rn == 1)
    {
      if ((rl[0].x == 0) && (rl[0].y == 0) &&
	  (rl[0].width == ewin->w) && (rl[0].height == ewin->h))
	{
	  if (rl)
	    XFree(rl);
	  return;
	}
    }
  mask = XCreatePixmap(disp, root.win, w, h, 1);
  gcm = XCreateGC(disp, mask, 0, &gcv);
  gcv.subwindow_mode = IncludeInferiors;
  gc = XCreateGC(disp, root.win, GCSubwindowMode, &gcv);

  XSetForeground(disp, gcm, 1);
  XFillRectangle(disp, mask, gcm, 0, 0, w, h);
  if (rl)
    {
      XSetForeground(disp, gcm, 0);
      for (i = 0; i < rn; i++)
	XFillRectangle(disp, mask, gcm, rl[i].x, rl[i].y, rl[i].width, rl[i].height);
      XFree(rl);
    }
  XSetClipMask(disp, gc, mask);
  XSetClipOrigin(disp, gc, x, y);
  XCopyArea(disp, pmap, root.win, gc, x, y, w, h, x, y);
  XFreePixmap(disp, mask);
  XFreeGC(disp, gc);
  XFreeGC(disp, gcm);
}

void
EBlendImage(EWin * ewin,
	    XImage * xim1, XShmSegmentInfo ** shminfo1,
	    XImage * xim2, XShmSegmentInfo ** shminfo2)
{
  XGCValues           gcv;
  GC                  gc;
  int                 i, j, rn, ord;
  int                 x, y, w, h, ww, hh;
  XRectangle         *rl;

  if ((ewin->sticky) || (ewin->floating))
    {
      x = ewin->x;
      y = ewin->y;
    }
  else
    {
      x = ewin->x + desks.desk[ewin->desktop].x;
      y = ewin->y + desks.desk[ewin->desktop].y;
    }
  w = ewin->w;
  h = ewin->h;

  rl = NULL;
  rl = XShapeGetRectangles(disp, ewin->win, ShapeBounding, &rn, &ord);
  gcv.subwindow_mode = IncludeInferiors;
  gc = XCreateGC(disp, root.win, GCSubwindowMode, &gcv);
  if (rl)
    {
      XSetClipRectangles(disp, gc, x, y, rl, rn, ord);
      XFree(rl);
    }

  if ((xim1) && (xim2))
    {
      ww = 0;
      hh = 0;
      if (xim1->width < xim2->width)
	ww = xim1->width;
      else
	ww = xim2->width;
      if (xim1->height < xim2->height)
	hh = xim1->height;
      else
	hh = xim2->height;
      switch (xim1->bits_per_pixel)
	{
	case 32:
	  for (j = 0; j < hh; j++)
	    {
	      unsigned int       *ptr1, *ptr2;

	      ptr1 = (unsigned int *)(xim1->data + (j * xim1->bytes_per_line));
	      ptr2 = (unsigned int *)(xim2->data + (j * xim2->bytes_per_line));
	      for (i = 0; i < ww; i++)
		{
		  unsigned int        p1, p2;

		  p1 = *ptr1++;
		  p2 = *ptr2;
		  *ptr2++ = ((p1 >> 1) & 0x7f7f7f7f) + ((p2 >> 1) & 0x7f7f7f7f) +
		    (p1 & p2 & 0x01010101);
		}
	    }
	  break;
	case 16:
	  if (id->x.render_depth != 15)
	    {
	      for (j = 0; j < hh; j++)
		{
		  unsigned short     *ptr1, *ptr2;

		  ptr1 = (unsigned short *)(xim1->data + (j * xim1->bytes_per_line));
		  ptr2 = (unsigned short *)(xim2->data + (j * xim2->bytes_per_line));
		  for (i = 0; i < ww; i++)
		    {
		      unsigned short      p1, p2;

		      p1 = *ptr1++;
		      p2 = *ptr2;
/** a + b / 2 (a = RGBA, b = RGBA)*/
		      /* ((a >> 1) & 0x7f7f7f7f) + 
		       * ((b >> 1) & 0x7f7f7f7f) + 
		       * (a & b & 0x01010101) */
		      *ptr2++ = ((p1 >> 1) & ((0x78 << 8) | (0x7c << 3) | (0x78 >> 3))) +
			((p2 >> 1) & ((0x78 << 8) | (0x7c << 3) | (0x78 >> 3))) +
			(p1 & p2 & ((0x1 << 11) | (0x1 << 5) | (0x1)));
		    }
		}
	    }
	  else
	    {
	      for (j = 0; j < hh; j++)
		{
		  unsigned short     *ptr1, *ptr2;

		  ptr1 = (unsigned short *)(xim1->data + (j * xim1->bytes_per_line));
		  ptr2 = (unsigned short *)(xim2->data + (j * xim2->bytes_per_line));
		  for (i = 0; i < ww; i++)
		    {
		      unsigned short      p1, p2;

		      p1 = *ptr1++;
		      p2 = *ptr2;
/** a + b / 2 (a = RGBA, b = RGBA)*/
		      /* ((a >> 1) & 0x7f7f7f7f) + 
		       * ((b >> 1) & 0x7f7f7f7f) + 
		       * (a & b & 0x01010101) */
		      *ptr2++ = ((p1 >> 1) & ((0x78 << 7) | (0x78 << 2) | (0x78 >> 3))) +
			((p2 >> 1) & ((0x78 << 7) | (0x78 << 2) | (0x78 >> 3))) +
			(p1 & p2 & ((0x1 << 10) | (0x1 << 5) | (0x1)));
		    }
		}
	    }
	  break;
	default:
	  for (j = 0; j < hh; j++)
	    {
	      unsigned char      *ptr1, *ptr2;

	      ptr1 = (unsigned char *)(xim1->data + (j * xim1->bytes_per_line));
	      ptr2 = (unsigned char *)(xim2->data + (j * xim2->bytes_per_line));
	      if (j & 0x1)
		{
		  ptr1++;
		  ptr2++;
		}
	      for (i = 0; i < ww; i += 2)
		{
		  unsigned char       p1;

		  p1 = *ptr1;
		  ptr1 += 2;
		  *ptr2 = p1;
		  ptr2 += 2;
		}
	    }
	  break;
	}
    }
  else if ((!xim1) && (xim2))
    {
      xim2 = xim1;
    }
  if (xim2)
    {
      if (*shminfo2)
	XShmPutImage(disp, root.win, gc, xim2, 0, 0, x, y, w, h, False);
      else
	XPutImage(disp, root.win, gc, xim2, 0, 0, x, y, w, h);
    }

  XSync(disp, False);
  XFreeGC(disp, gc);
  shminfo1 = NULL;
}

#include <X11/bitmaps/flipped_gray>
#include <X11/bitmaps/gray>
#include <X11/bitmaps/gray3>

void
DrawEwinShape(EWin * ewin, int md, int x, int y, int w, int h, char firstlast)
{
  static GC           gc = 0;
  XGCValues           gcv;
  int                 x1, y1, w1, h1, i, j, pw, ph;
  static Pixmap       b1 = 0, b2 = 0, b3 = 0;
  static Pixmap       parea = 0;
  static XImage      *ewin_xim = NULL, *bg_xim = NULL;
  static XShmSegmentInfo *ewin_shminfo = NULL, *bg_shminfo = NULL;
  static int          last_md = -1;
  static char         changemode = 0;

  EDBUG(4, "DrawEwinShape");
  if ((mode.mode == MODE_RESIZE) ||
      (mode.mode == MODE_RESIZE_H) ||
      (mode.mode == MODE_RESIZE_V))
    {
      w1 = ewin->client.w;
      h1 = ewin->client.h;
      ewin->client.w = w;
      ewin->client.h = h;
      ICCCM_MatchSize(ewin);
      i = (x - ewin->x) / ewin->client.w_inc;
      j = (y - ewin->y) / ewin->client.h_inc;
      x = ewin->x + (i * ewin->client.w_inc);
      y = ewin->y + (j * ewin->client.h_inc);
      ewin->client.w = w1;
      ewin->client.h = h1;
    }
  if ((md == 5)
      && ((mode.mode == MODE_RESIZE) ||
	  (mode.mode == MODE_RESIZE_H) ||
	  (mode.mode == MODE_RESIZE_V) ||
	  ((last_md != 5) && (firstlast == 2))))
    md = 3;
  last_md = md;
  if (md == 5)
    {
      if (((x + desks.desk[ewin->desktop].x < 0) ||
	   (x + desks.desk[ewin->desktop].x + ewin->w > root.w) ||
	   (y + desks.desk[ewin->desktop].y < 0) ||
	   (y + desks.desk[ewin->desktop].y + ewin->h > root.h)) ||
	  (changemode))
	{
	  if ((firstlast == 1) && (!changemode))
	    {
	      if (parea)
		EPastePixmap(root.win, parea,
			     ewin->x + desks.desk[ewin->desktop].x,
			     ewin->y + desks.desk[ewin->desktop].y,
			     ewin->w, ewin->h);
	      if (parea)
		XFreePixmap(disp, parea);
	      DrawEwinShape(ewin, 3, ewin->x, ewin->y, w, h, 3);
	    }
	  md = 3;
	  changemode = 1;
	  if (firstlast == 5)
	    EDBUG_RETURN_;
	  if (firstlast == 4)
	    firstlast = 0;
	}
      if ((firstlast == 2) && (changemode))
	changemode = 0;
    }
  pw = w;
  ph = h;
  switch (md)
    {
    case 0:
      MoveResizeEwin(ewin, x, y, w, h);
      break;
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
      if (!b1)
	b1 = XCreateBitmapFromData(disp, root.win, flipped_gray_bits,
				   flipped_gray_width,
				   flipped_gray_height);
      if (!b2)
	b2 = XCreateBitmapFromData(disp, root.win, gray_bits,
				   gray_width,
				   gray_height);
      if (!b3)
	b3 = XCreateBitmapFromData(disp, root.win, gray3_bits,
				   gray3_width,
				   gray3_height);
      x1 = ewin->x + desks.desk[ewin->desktop].x;
      y1 = ewin->y + desks.desk[ewin->desktop].y;
      w1 = ewin->w - (ewin->border->border.left + ewin->border->border.right);
      h1 = ewin->h - (ewin->border->border.top + ewin->border->border.bottom);
      ewin->x = x;
      ewin->y = y;
      ewin->reqx = x;
      ewin->reqy = y;
      x = ewin->x + desks.desk[ewin->desktop].x;
      y = ewin->y + desks.desk[ewin->desktop].y;
      if ((w != ewin->client.w) || (h != ewin->client.h))
	{
	  ewin->client.w = w;
	  ewin->client.h = h;
	  ICCCM_MatchSize(ewin);
	  ewin->w = ewin->client.w + ewin->border->border.left + ewin->border->border.right;
	  ewin->h = ewin->client.h + ewin->border->border.top + ewin->border->border.bottom;
	}
      w = ewin->w - (ewin->border->border.left + ewin->border->border.right);
      h = ewin->h - (ewin->border->border.top + ewin->border->border.bottom);
      fprintf(stderr, "%i %i | %i %i\n", ewin->w, ewin->h, w, h);
      fflush(stderr);
      if (!gc)
	{
	  gcv.function = GXxor;
	  gcv.foreground = WhitePixel(disp, root.scr);
	  if (gcv.foreground == 0)
	    gcv.foreground = BlackPixel(disp, root.scr);
	  gcv.subwindow_mode = IncludeInferiors;
	  gc = XCreateGC(disp, root.win, GCFunction | GCForeground | GCSubwindowMode, &gcv);
	}
#define DO_DRAW_MODE_1(aa, bb, cc, dd) \
      XDrawLine(disp, root.win, gc, aa, 0, aa, root.h); \
      XDrawLine(disp, root.win, gc, \
		aa + cc + ewin->border->border.left + \
		ewin->border->border.right - 1, 0, \
		aa + cc + ewin->border->border.left + \
		ewin->border->border.right - 1, root.h); \
      XDrawLine(disp, root.win, gc, 0, bb, root.w, bb); \
      XDrawLine(disp, root.win, gc, 0, \
		bb + dd + ewin->border->border.top + \
		ewin->border->border.bottom - 1, root.w, \
		bb + dd + ewin->border->border.top + \
		ewin->border->border.bottom - 1); \
      XDrawRectangle(disp, root.win, gc, aa + ewin->border->border.left, \
		     bb + ewin->border->border.top, cc - 1, dd - 1);

#define DO_DRAW_MODE_2(aa, bb, cc, dd) \
      XDrawRectangle(disp, root.win, gc, aa, bb, \
                     cc + ewin->border->border.left + \
                     ewin->border->border.right - 1, \
                     dd + ewin->border->border.top + \
                     ewin->border->border.bottom - 1); \
      XDrawRectangle(disp, root.win, gc, aa + ewin->border->border.left, \
		     bb + ewin->border->border.top, cc - 1, dd - 1);

#define DO_DRAW_MODE_3(aa, bb, cc, dd) \
      XSetFillStyle(disp, gc, FillStippled); \
      XSetStipple(disp, gc, b2); \
      if ((cc + ewin->border->border.left + ewin->border->border.right > 0) && \
          (ewin->border->border.top > 0)) \
      XFillRectangle(disp, root.win, gc, aa, bb, \
                     cc + ewin->border->border.left + \
                     ewin->border->border.right, \
                     ewin->border->border.top); \
      if ((cc + ewin->border->border.left + ewin->border->border.right > 0) && \
          (ewin->border->border.bottom > 0)) \
      XFillRectangle(disp, root.win, gc, aa, bb + dd + \
                     ewin->border->border.top, \
                     cc + ewin->border->border.left + \
                     ewin->border->border.right, \
                     ewin->border->border.bottom); \
      if ((dd > 0) && (ewin->border->border.left > 0)) \
      XFillRectangle(disp, root.win, gc, aa, bb + ewin->border->border.top, \
                     ewin->border->border.left, \
                     dd); \
      if ((dd > 0) && (ewin->border->border.right > 0)) \
      XFillRectangle(disp, root.win, gc, aa + cc + ewin->border->border.left, \
                     bb + ewin->border->border.top, \
                     ewin->border->border.right, \
                     dd); \
      XSetStipple(disp, gc, b3); \
      if ((cc > 0) && (dd > 0)) \
        XFillRectangle(disp, root.win, gc, aa + ewin->border->border.left, \
  		       bb + ewin->border->border.top, cc - 1, dd - 1);

#define DO_DRAW_MODE_4(aa, bb, cc, dd) \
      XSetFillStyle(disp, gc, FillStippled); \
      XSetStipple(disp, gc, b2); \
      XFillRectangle(disp, root.win, gc, aa, bb, \
                     cc + ewin->border->border.left + \
                     ewin->border->border.right, \
                     dd + ewin->border->border.top + \
                     ewin->border->border.bottom);
      if (md == 1)
	{
	  if (firstlast > 0)
	    {
	      DO_DRAW_MODE_1(x1, y1, w1, h1);
	    }
	  if (firstlast < 2)
	    {
	      DO_DRAW_MODE_1(x, y, w, h);
	    }
	}
      else if (md == 2)
	{
	  if (firstlast > 0)
	    {
	      DO_DRAW_MODE_2(x1, y1, w1, h1);
	    }
	  if (firstlast < 2)
	    {
	      DO_DRAW_MODE_2(x, y, w, h);
	    }
	}
      else if (md == 3)
	{
	  if (firstlast > 0)
	    {
	      DO_DRAW_MODE_3(x1, y1, w1, h1);
	    }
	  if (firstlast < 2)
	    {
	      DO_DRAW_MODE_3(x, y, w, h);
	    }
	}
      else if (md == 4)
	{
	  if (firstlast > 0)
	    {
	      DO_DRAW_MODE_4(x1, y1, w1, h1);
	    }
	  if (firstlast < 2)
	    {
	      DO_DRAW_MODE_4(x, y, w, h);
	    }
	}
      else if (md == 5)
	{
	  if (firstlast == 0)
	    {
	      if (parea)
		XFreePixmap(disp, parea);
	      if (ewin_xim)
		EDestroyImage(ewin_xim, &ewin_shminfo);
	      if (bg_xim)
		EDestroyImage(bg_xim, &bg_shminfo);
	      parea = 0;
	      ewin_xim = NULL;
	      bg_xim = NULL;
	      ewin_shminfo = NULL;
	      bg_shminfo = NULL;
	      parea = EInitPixmap(root.win);
	      EFillPixmap(root.win, parea, x1, y1, ewin->w, ewin->h);
	      ewin_xim = EInitImage(&ewin_shminfo, ewin);
	      bg_xim = EInitImage(&bg_shminfo, ewin);
	      ewin_xim = EFillImage(ewin_xim, &ewin_shminfo, ewin, root.win);
	      bg_xim = EFillImage(bg_xim, &bg_shminfo, ewin, parea);
	      EBlendImage(ewin,
			  ewin_xim, &ewin_shminfo,
			  bg_xim, &bg_shminfo);
	    }
	  else if (firstlast == 1)
	    {
	      int                 dx, dy, wt, ht;
	      int                 adx, ady;

	      dx = x - x1;
	      dy = y - y1;
	      if (dx < 0)
		adx = -dx;
	      else
		adx = dx;
	      if (dy < 0)
		ady = -dy;
	      else
		ady = dy;
	      wt = ewin->w;
	      ht = ewin->h;
	      if ((adx <= wt) && (ady <= ht))
		{
		  if (dx < 0)
		    EFillPixmap(root.win, parea, x, y, -dx, ht);
		  else if (dx > 0)
		    EFillPixmap(root.win, parea, x + wt - dx, y, dx, ht);
		  if (dy < 0)
		    EFillPixmap(root.win, parea, x, y, wt, -dy);
		  else if (dy > 0)
		    EFillPixmap(root.win, parea, x, y + ht - dy, wt, dy);
		}
	      else
		EFillPixmap(root.win, parea, x, y, wt, ht);
	      bg_xim = EFillImage(bg_xim, &bg_shminfo, ewin, parea);
	      if ((adx <= wt) && (ady <= ht))
		{
		  EBlendImage(ewin,
			      ewin_xim, &ewin_shminfo,
			      bg_xim, &bg_shminfo);
		  if (dx > 0)
		    EPastePixmap(root.win, parea, x1, y1, dx, ht);
		  else if (dx < 0)
		    EPastePixmap(root.win, parea, x1 + wt + dx, y1, -dx, ht);
		  if (dy > 0)
		    EPastePixmap(root.win, parea, x1, y1, wt, dy);
		  else if (dy < 0)
		    EPastePixmap(root.win, parea, x1, y1 + ht + dy, wt, -dy);
		}
	      else
		{
		  EPastePixmap(root.win, parea, x1, y1, wt, ht);
		  EBlendImage(ewin,
			      ewin_xim, &ewin_shminfo,
			      bg_xim, &bg_shminfo);
		}
	      EBlendRemoveShape(ewin, parea);
	    }
	  else if (firstlast == 2)
	    {
	      EPastePixmap(root.win, parea, x1, y1, ewin->w, ewin->h);
	      if (parea)
		XFreePixmap(disp, parea);
	      if (ewin_xim)
		EDestroyImage(ewin_xim, &ewin_shminfo);
	      if (bg_xim)
		EDestroyImage(bg_xim, &bg_shminfo);
	      parea = 0;
	      ewin_xim = NULL;
	      bg_xim = NULL;
	      ewin_shminfo = NULL;
	      bg_shminfo = NULL;
	    }
	  else if (firstlast == 3)
	    {
	      EPastePixmap(root.win, parea, x, y, ewin->w, ewin->h);
	      if (parea)
		XFreePixmap(disp, parea);
	      parea = 0;
	    }
	  else if (firstlast == 4)
	    {
	      int                 wt, ht;

	      wt = ewin->w;
	      ht = ewin->h;
	      parea = EInitPixmap(root.win);
	      EFillPixmap(root.win, parea, x, y, wt, ht);
	      bg_xim = EFillImage(bg_xim, &bg_shminfo, ewin, parea);
	      EBlendImage(ewin,
			  ewin_xim, &ewin_shminfo,
			  bg_xim, &bg_shminfo);
	    }
	  else if (firstlast == 5)
	    {
	      if (parea)
		XFreePixmap(disp, parea);
	      parea = 0;
	    }
	}
      if (firstlast == 2)
	{
	  if (ewin->shaded)
	    MoveEwin(ewin, ewin->x, ewin->y);
	  else
	    MoveResizeEwin(ewin, ewin->x, ewin->y, pw, ph);
	  XFreeGC(disp, gc);
	  gc = 0;
	}
      break;
    default:
      break;
    }
  EDBUG_RETURN_;
}

ImlibImage         *
ELoadImage(char *file)
{
  EDBUG(5, "ELoadImage");
  EDBUG_RETURN(ELoadImageImlibData(id, file));
}

ImlibImage         *
ELoadImageImlibData(ImlibData * imd, char *file)
{
  ImlibImage         *im;
  char               *f;

  EDBUG(5, "ELoadImageImlibData");
  f = FindFile(file);
  if (f)
    {
      im = Imlib_load_image(imd, f);
      Efree(f);
      EDBUG_RETURN(im);
    }
  EDBUG_RETURN(NULL);
}

void
PropagateShapes(Window win)
{
  Window              rt, par, *list = NULL;
  int                 k, i, num = 0, num_rects = 0, rn = 0, ord;
  int                 x, y;
  unsigned int        ww, hh, w, h, d;
  XRectangle         *rects = NULL, *rl = NULL;
  XWindowAttributes   att;

  EDBUG(6, "PropagateShapes");
  if (queue_up)
    {
      DrawQueue          *dq;

      dq = Emalloc(sizeof(DrawQueue));
      dq->win = win;
      dq->iclass = NULL;
      dq->w = 0;
      dq->h = 0;
      dq->active = 0;
      dq->sticky = 0;
      dq->state = 0;
      dq->expose = 0;
      dq->tclass = NULL;
      dq->text = NULL;
      dq->shape_propagate = 1;
      AddItem(dq, "DRAW", dq->win, LIST_TYPE_DRAW);
      EDBUG_RETURN_;
    }

  XGetGeometry(disp, win, &rt, &x, &y, &w, &h, &d, &d);
  if ((w <= 0) || (h <= 0))
    EDBUG_RETURN_;

  ww = w;
  hh = h;

  XQueryTree(disp, win, &rt, &par, &list, (unsigned int *)&num);
  if (list)
    {
      /* go through all child windows and create/inset spans */
      for (i = 0; i < num; i++)
	{
	  XGetWindowAttributes(disp, list[i], &att);
	  x = att.x;
	  y = att.y;
	  w = att.width;
	  h = att.height;
	  if ((att.class == InputOutput) && (att.map_state != IsUnmapped))
	    {
	      rl = NULL;
	      rl = XShapeGetRectangles(disp, list[i], ShapeBounding, &rn, &ord);
	      if (rl)
		{
		  num_rects += rn;
		  rects = Erealloc(rects, num_rects * sizeof(XRectangle));
		  /* go through all clip rects in thsi window's shape */
		  for (k = 0; k < rn; k++)
		    {
		      /* for each clip rect, add it to the rect list */
		      rects[num_rects - rn + k].x = x + rl[k].x;
		      rects[num_rects - rn + k].y = y + rl[k].y;
		      rects[num_rects - rn + k].width = rl[k].width;
		      rects[num_rects - rn + k].height = rl[k].height;
		    }
		  XFree(rl);
		}
	    }
	}
      /* set the rects as the shape mask */
      if (rects)
	{
	  XShapeCombineRectangles(disp, win, ShapeBounding, 0, 0, rects, num_rects,
				  ShapeSet, Unsorted);
	  Efree(rects);
	  rl = NULL;
	  rl = XShapeGetRectangles(disp, win, ShapeBounding, &rn, &ord);
	  if (rl)
	    {
	      if (rn < 1)
		XShapeCombineMask(disp, win, ShapeBounding, 0, 0,
				  None, ShapeSet);
	      else if (rn == 1)
		{
		  if ((rl[0].x == 0) && (rl[0].y == 0) &&
		      (rl[0].width == ww) && (rl[0].height == hh))
		    XShapeCombineMask(disp, win, ShapeBounding, 0, 0,
				      None, ShapeSet);
		}
	      XFree(rl);
	    }
	}
      XFree(list);
    }
  EDBUG_RETURN_;
}
