#include "E.h"

EWin               *
FindEwinByBase(Window win)
{
  EWin              **ewins;
  EWin               *ewin;
  int                 i, num;

  EDBUG(6, "FindEwinByBase");
  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      if (win == ewins[i]->win)
	{
	  ewin = ewins[i];
	  Efree(ewins);
	  EDBUG_RETURN(ewin);
	}
    }
  if (ewins)
    Efree(ewins);
  EDBUG_RETURN(NULL);
}

EWin               *
FindEwinByChildren(Window win)
{
  EWin               *ewin;
  EWin              **ewins;
  int                 i, j, num;

  EDBUG(6, "FindEwinByChildren");

  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      if (win == ewins[i]->client.win)
	{
	  ewin = ewins[i];
	  Efree(ewins);
	  EDBUG_RETURN(ewin);
	}
      else if (win == ewins[i]->win_container)
	{
	  ewin = ewins[i];
	  Efree(ewins);
	  EDBUG_RETURN(ewin);
	}
      else if (win == ewins[i]->win_event)
	{
	  ewin = ewins[i];
	  Efree(ewins);
	  EDBUG_RETURN(ewin);
	}
      else
	{
	  for (j = 0; j < ewins[i]->border->num_winparts; j++)
	    if (win == ewins[i]->bits[j].win)
	      {
		ewin = ewins[i];
		Efree(ewins);
		EDBUG_RETURN(ewin);
	      }
	}
    }
  if (ewins)
    Efree(ewins);
  EDBUG_RETURN(NULL);
}

EWin               *
FindEwinByDecoration(Window win)
{
  EWin               *ewin;
  EWin              **ewins;
  int                 i, j, num;

  EDBUG(6, "FindEwinByDecoration");

  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      for (j = 0; j < ewins[i]->border->num_winparts; j++)
	{
	  if (win == ewins[i]->bits[j].win)
	    {
	      ewin = ewins[i];
	      Efree(ewins);
	      EDBUG_RETURN(ewin);
	    }
	}
    }
  if (ewins)
    Efree(ewins);
  EDBUG_RETURN(NULL);
}

Button             *
FindButton(Window win)
{
  Button             *b;
  Button            **buttons;
  int                 i, num;

  EDBUG(6, "FindButton");

  buttons = (Button **) ListItemType(&num, LIST_TYPE_BUTTON);
  for (i = 0; i < num; i++)
    {
      if ((win == buttons[i]->win) || (win == buttons[i]->inside_win) ||
	  (win == buttons[i]->event_win))
	{
	  b = buttons[i];
	  Efree(buttons);
	  EDBUG_RETURN(b);
	}
    }
  if (buttons)
    Efree(buttons);
  EDBUG_RETURN(NULL);
}

ActionClass        *
FindActionClass(Window win)
{
  Button             *b;
  EWin               *ewin;
  int                 i;

  EDBUG(6, "FindActionClass");
  b = FindButton(win);
  if (b)
    EDBUG_RETURN(b->aclass);
  ewin = FindEwinByDecoration(win);
  if (ewin)
    {
      for (i = 0; i < ewin->border->num_winparts; i++)
	if (win == ewin->bits[i].win)
	  EDBUG_RETURN(ewin->border->part[i].aclass);
    }
  EDBUG_RETURN(NULL);
}

Menu               *
FindMenuItem(Window win, MenuItem ** mi)
{
  Menu               *menu;
  Menu              **menus;
  int                 i, j, num;

  EDBUG(6, "FindMenuItem");

  menus = (Menu **) ListItemType(&num, LIST_TYPE_MENU);
  for (i = 0; i < num; i++)
    {
      for (j = 0; j < menus[i]->num; j++)
	{
	  if ((win == menus[i]->items[j]->win) ||
	      (win == menus[i]->items[j]->icon_win))
	    {
	      *mi = menus[i]->items[j];
	      menu = menus[i];
	      Efree(menus);
	      EDBUG_RETURN(menu);
	    }
	}
    }
  if (menus)
    Efree(menus);
  EDBUG_RETURN(NULL);
}

Menu               *
FindMenu(Window win)
{
  Menu               *menu;
  Menu              **menus;
  int                 i, num;

  EDBUG(6, "FindMenu");
  menus = (Menu **) ListItemType(&num, LIST_TYPE_MENU);
  for (i = 0; i < num; i++)
    {
      if (menus[i]->win == win)
	{
	  menu = menus[i];
	  Efree(menus);
	  EDBUG_RETURN(menu);
	}
    }
  if (menus)
    Efree(menus);
  EDBUG_RETURN(NULL);
}

EWin               *
FindEwinByMenu(Menu * m)
{
  EWin               *ewin;
  EWin              **ewins;
  int                 i, num;

  EDBUG(6, "FindEwinByMenu");
  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      if (ewins[i]->menu == m)
	{
	  ewin = ewins[i];
	  Efree(ewins);
	  EDBUG_RETURN(ewin);
	}
    }
  if (ewins)
    Efree(ewins);
  EDBUG_RETURN(NULL);
}
