#include "E.h"
#include <audiofile.h>

typedef struct _WAVFormatChunk
  {
    char                chunkID[4];
    int                 chunkSize;

    short               wFormatTag;
    unsigned short      wChannels;
    unsigned int        dwSamplesPerSec;
    unsigned int        dwAvgBytesPerSec;
    unsigned short      wBlockAlign;
    unsigned short      wBitsPerSample;
  }
WAVFormatChunk;

#ifdef WORDS_BIGENDIAN
#define SWAP_SHORT( x ) x = ( ( x & 0x00ff ) << 8 ) | ( ( x >> 8 ) & 0x00ff )
#define SWAP_LONG( x ) x = ( ( ( x & 0x000000ff ) << 24 ) |\
( ( x & 0x0000ff00 ) << 8 ) |\
( ( x & 0x00ff0000 ) >> 8 ) |\
( ( x & 0xff000000 ) >> 24 ) )
#endif

Sample             *
LoadWav(char *file)
{
  AFfilehandle        in_file;
  char                str[FILEPATH_LEN_MAX];
  char                found = 0;
  Sample             *s;
  int                 in_format, in_width, in_channels, frame_count;
  double              in_rate;
  int                 bytes_per_frame, frames_read;

  /* int                 count; */

  EDBUG(5, "LoadWav");
#ifdef HAVE_LIBESD
  if (exists(file))
    {
      found = 1;
      Esnprintf(str, sizeof(str), "%s", file);
    }
  if (!found)
    {
      if (themepath[0])
	{
	  Esnprintf(str, sizeof(str), "%s/%s", themepath, file);
	  if (exists(str))
	    {
	      found = 1;
	    }
	}
      if (!found)
	{
	  Esnprintf(str, sizeof(str), "%s/.enlightenment/%s", homedir(getuid()), file);
	  if (!exists(str))
	    {
	      Esnprintf(str, sizeof(str), "%s/config/%s", ENLIGHTENMENT_ROOT, file);
	      if (!exists(str))
		{
		  Esnprintf(str, sizeof(str), "Warning!  Enlightenment was unable "
			    "to load the\nfollowing sound file:\n%s\n"
			    "Enlightenment will continue to operate, but you\n"
			    "may wish to check your configuration settings.\n",
			    file);
		  Alert(str);
		  Esnprintf(str, sizeof(str), "%s/config/%s", ENLIGHTENMENT_ROOT, file);
		}
	    }
	}
    }
  in_file = afOpenFile(str, "r", NULL);
  if (!in_file)
    EDBUG_RETURN(NULL);
  s = Emalloc(sizeof(Sample));
  if (!s)
    {
      afCloseFile(in_file);
      EDBUG_RETURN(NULL);
    }
  frame_count = afGetFrameCount(in_file, AF_DEFAULT_TRACK);
  in_channels = afGetChannels(in_file, AF_DEFAULT_TRACK);
  in_rate = afGetRate(in_file, AF_DEFAULT_TRACK);
  afGetSampleFormat(in_file, AF_DEFAULT_TRACK, &in_format, &in_width);
  afSetVirtualByteOrder(in_file, AF_DEFAULT_TRACK, AF_BYTEORDER_LITTLEENDIAN);

  s->rate = 44100;
  s->format = ESD_STREAM | ESD_PLAY;
  s->samples = 0;
  s->data = NULL;
  s->id = 0;

  if (in_width == 8)
    s->format |= ESD_BITS8;
  else if (in_width == 16)
    s->format |= ESD_BITS16;
  bytes_per_frame = (in_width * in_channels) / 8;
  if (in_channels == 1)
    s->format |= ESD_MONO;
  else if (in_channels == 2)
    s->format |= ESD_STEREO;
  s->rate = (int)in_rate;

  s->samples = frame_count * bytes_per_frame;
  s->data = Emalloc(frame_count * bytes_per_frame);
  frames_read = afReadFrames(in_file, AF_DEFAULT_TRACK, s->data, s->samples);
  afCloseFile(in_file);
  EDBUG_RETURN(s);
#endif
  EDBUG_RETURN(NULL);
}

void
SoundPlay(Sample * s)
{
  int                 size;

  int                 confirm = 0;

  EDBUG(5, "SoundPlay");
#ifdef HAVE_LIBESD
  if ((sound_fd < 0) || (!mode.sound) || (!s))
    EDBUG_RETURN_;
  size = s->samples;

  /* #if 1 for esound-0.2.3, #if 0 for esound > 0.2.3 */
#if 0
  if (!s->id)
    {
      int                 fd = -1;

      fd = esd_open_sound("Enlightenment");
      if (fd >= 0)
	{
	  if (s->data)
	    {
	      s->id = esd_sample_cache(fd, s->format, s->rate, size);
	      write(fd, s->data, size * 2);
	      Efree(s->data);
	      s->data = NULL;
	    }
	  close(fd);
	}
    }
  esd_sample_play(sound_fd, s->id);
  fsync(sound_fd);
#else
  if (!s->id)
    {
      if (sound_fd >= 0)
	{
	  if (s->data)
	    {
	      /* "name" of all samples is currently "E", should be name of sound 
	       * file, or event type, for later identification */
	      s->id = esd_sample_cache(sound_fd, s->format, s->rate,
				       size, "E");
	      write(sound_fd, s->data, size);
	      confirm = esd_confirm_sample_cache(sound_fd);
	      if (s->id <= 0 || confirm != s->id)
		{
		  printf("SoundPlay: error caching sample <%d>!\n", s->id);
		  s->id = 0;
		}
	      Efree(s->data);
	      s->data = NULL;
	    }
	}
    }
  if (s->id > 0)
    esd_sample_play(sound_fd, s->id);

  fsync(sound_fd);
#endif

#endif
  EDBUG_RETURN_;
}

SoundClass         *
CreateSoundClass(char *name, char *file)
{
  SoundClass         *sclass;

  EDBUG(6, "CreateSoundClass");
  sclass = Emalloc(sizeof(SoundClass));
  if (!sclass)
    EDBUG_RETURN(NULL);
  sclass->name = duplicate(name);
  sclass->file = duplicate(file);
  sclass->sample = NULL;
  EDBUG_RETURN(sclass);
}

void
ApplySclass(SoundClass * sclass)
{
  char               *f;

  EDBUG(4, "ApplySclass");
  if (!sclass)
    EDBUG_RETURN_;
#ifdef HAVE_LIBESD
  if (!sclass->sample)
    {
      f = FindFile(sclass->file);
      if (f)
	{
	  sclass->sample = LoadWav(f);
	  Efree(f);
	}
    }
#endif
  SoundPlay(sclass->sample);
  EDBUG_RETURN_;
}

void
DestroySample(Sample * s)
{
  EDBUG(5, "DestroySample");

#ifdef HAVE_LIBESD
  if ((s->id) && (sound_fd >= 0))
    {
/*      Why the hell is this symbol not in esd? */
/*      it's in esd.h - evil evil evil */
/*      esd_sample_kill(sound_fd,s->id); */
      esd_sample_free(sound_fd, s->id);
    }
#endif
  if (s->data)
    Efree(s->data);
  if (s)
    Efree(s);
  EDBUG_RETURN_;
}

void
DestroySclass(SoundClass * sclass)
{
  if (!sclass)
    EDBUG_RETURN_;

  EDBUG(5, "DestroySclass");
  if (sclass->name)
    Efree(sclass->name);
  if (sclass->file)
    Efree(sclass->file);
  if (sclass->sample)
    DestroySample(sclass->sample);
  Efree(sclass);
  EDBUG_RETURN_;
}

void
SoundInit()
{
  int                 fd;

  EDBUG(5, "SoundInit");
#ifdef HAVE_LIBESD
  sound_fd = -1;
  if (!mode.sound)
    EDBUG_RETURN_;
  fd = esd_open_sound(NULL);
  if (fd >= 0)
    sound_fd = fd;
  else
    {
      execApplication("esd");
      sleep(2);
      fd = esd_open_sound(NULL);
      if (fd >= 0)
	sound_fd = fd;
      else
	{
	  Alert("Audio has been turned on for Enlightenment, but\n"
		"Enlightenment is unable to find the ESD sound server or\n"
		"run it.\n"
		"As a result Audio will now be turned off.\n"
		"You can safely hit Left Button for OK and Enlightenment\n"
		"will continue to run without audio.\n"
		"\n"
		"Hitting middle or right button to Restart or Quit is\n"
		"also an option if you believe ESD should be runing but\n"
		"isn't and you wish to fix it.\n"
		"It could be the esd executable is not in the PATH for\n"
		"your shell.\n"
		"\n"
		"If audio is not meant to work on your system, then\n"
		"please disable it now to avoid messages like this in\n"
		"future\n");
	  mode.sound = 0;
	}
    }
#else
  mode.sound = 0;
#endif
  EDBUG_RETURN_;
}

void
SoundExit()
{
  SoundClass        **lst;
  int                 num, i;

  EDBUG(6, "SoundExit");
  if (sound_fd >= 0)
    {
      lst = (SoundClass **) ListItemType(&num, LIST_TYPE_SCLASS);
      if (lst)
	{
	  for (i = 0; i < num; i++)
	    {
	      if (lst[i]->sample)
		DestroySample(lst[i]->sample);
	      lst[i]->sample = NULL;
	    }
	  Efree(lst);
	}
      close(sound_fd);
      sound_fd = -1;
    }
  EDBUG_RETURN_;
}
